// Constants
const NUMBER_BORDER_COLORS = 12;
const ANIMATION_INTERVAL = 40;
const HTML_NS = "http://www.w3.org/1999/xhtml";
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
const STATE_START = Components.interfaces.nsIWebProgressListener.STATE_START;
const STATE_STOP = Components.interfaces.nsIWebProgressListener.STATE_STOP;
const STATE_REDIRECTING= Components.interfaces.nsIWebProgressListener.STATE_REDIRECTING;
const STATE_TRANSFERRING = Components.interfaces.nsIWebProgressListener.STATE_TRANSFERRING;
const STATE_IS_DOCUMENT = Components.interfaces.nsIWebProgressListener.STATE_IS_DOCUMENT;
const STATE_IS_WINDOW = Components.interfaces.nsIWebProgressListener.STATE_IS_WINDOW;
const STATE_IS_REQUEST = Components.interfaces.nsIWebProgressListener.STATE_IS_REQUEST;
const RELOAD_FLAGS = Components.interfaces.nsIWebNavigation.LOAD_FLAGS_NONE;
const RELOAD_SKIP_CACHE_FLAGS = Components.interfaces.nsIWebNavigation.LOAD_FLAGS_BYPASS_PROXY | Components.interfaces.nsIWebNavigation.LOAD_FLAGS_BYPASS_CACHE;
const VK_SPACE = 32;
const IETAB_URL = "chrome://ietab/content/reloaded.html?url=";
const SHOWCASE_SIDEBAR = "viewShowcaseSidebar";
const SHOWCASETHISWINDOW_SIDEBAR = "viewShowcaseThisWindowSidebar";
const OPTIONS_URL = "chrome://showcase/content/pref-showcase.xul";

// Global vars
var globalMode;
var topWindow;
var targetFocusWindow;
var currentAnimations = new Array();
var loadingTabTitle;
var untitledTabTitle;
var closeTabTitle;
var showcasePopupTab;
var pendingThumbnails = new Array();
var windowCurrentColor = 1;
var thumbWidth, thumbHeight;
var closingShowcase = false;
var showcaseWindows = new Array();
var currentCursor = null;
var currentCursorWindow = null;
var showcaseOriginalTitle = null;
var usingSeaMonkey = false;
var tabOldStyle;
var showcaseOldBackgroundColor;

// Configuration parameters
var coloredThumbnailsForMultipleWindows;
var paintThumbnailsAfterLoading;
var paintSelectedTabsFirst;
var thumbnailActiveAnimation;
var animationFrames;
var thumbnailIdleSizeFactor;
var thumbnailActiveSizeFactor;
var windowAttributeMaximized;
var windowAttributeFullScreen;
var windowAttributeTranslucent;
var updateThumbnailWhenContentLoaded;
var updateThumbnailWhenContentChanges;
var limitContentChangeUpdates;
var limitContentChangeUpdatesTime;
var thumbnailShowLabels;
var thumbnailLabelStyle;
var leftClickAction;
var middleClickAction;
var tabCustomColors = false;
var tabCustomTitleColor;
var tabCustomBackgroundColor;
var showcaseCloseWhenBlur;
var showcaseCloseWhenSelect;
var showcaseCloseWhenZeroThumbnails;

// Observers
var windowObserver = {
  observe: function(subject,topic,data){
    switch(topic){
      case 'domwindowopened':
        delayedCreateContent();
        break;
      case 'domwindowclosed':
        if (unlinkWindow(subject)) {
          if ((showcaseWindows.length == 0) && showcaseCloseWhenZeroThumbnails) {
            closeWindow();
          } else {
            if (showcaseWindows.length == 1) {
              updateBorders(showcaseWindows[0]);
            }
            updateTitle();
            refreshContent();
          }
        }
        break;
    }
  }
};

// Mosaic vars
var mosaicRows;
var mosaicColumns;
var mosaicWidth;
var mosaicHeight;

// Zoom vars
var inZoomMode = false;
var zoomTargetThumbnail = null;
var zoomTargetPreviousCursorStyle = null;

// Variable to generate unique ids for our elements
var idGenerator = 0;

// Array helper functions
Array.prototype.removeAllElements = function(){
  this.splice(0, this.length);
}

Array.prototype.contains = function(target) {
  return (this.indexOf(target) > -1);
}

Array.prototype.removeElementAt = function(index) {
  this.splice(index, 1);
}

Array.prototype.insertElementAt = function(element, index) {
  this.splice(index, 0, element);
}

function startup() {
  usingSeaMonkey = (navigator.userAgent.indexOf("SeaMonkey") > -1);

  var closedEmbedded = false;

  // Guess if we want single window mode or multi window
  if (window.location.search) {
    var globalModeParameter = window.location.search.substr("?global=".length, window.location.search.length);
    globalMode = ((globalModeParameter == "true") || (globalModeParameter == "yes"));
  } else {
    globalMode = false;
  }

  // Set top window if needed
  topWindow = QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIWebNavigation).QueryInterface(Components.interfaces.nsIDocShellTreeItem).rootTreeItem.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindow);
  if ((topWindow.embeddedShowcaseWindow) && (topWindow.embeddedShowcaseWindow != window)) {
    closedEmbedded = true;
    topWindow.embeddedShowcaseWindow.closeWindow();
  }

  if (globalMode) {
    // Close any other showcase instance!
    var wm = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
    var windowIter = wm.getEnumerator('extension:showcase');
    if (windowIter.hasMoreElements()) {
      do {
        var currentWindow = windowIter.getNext();
        if (currentWindow != window) {
          currentWindow.closeWindow();
        }
      } while (windowIter.hasMoreElements());
    }

    // Close embedded
    windowIter = wm.getEnumerator('navigator:browser');
    while (windowIter.hasMoreElements()) {
      var currentWindow = windowIter.getNext();
      if (currentWindow.embeddedShowcaseWindow) {
        currentWindow.embeddedShowcaseWindow.closeWindow();
      }
    }
  } else {
    // Close any global showcase, or local one that targets "us"
    var wm = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
    var windowIter = wm.getEnumerator('extension:showcase');
    var parentWindow = getParentWindow();
    if (windowIter.hasMoreElements()) {
      do {
        var currentWindow = windowIter.getNext();
        if (currentWindow.globalMode || (currentWindow.showcaseWindows && (currentWindow.showcaseWindows.length > 0) && currentWindow.showcaseWindows[0] && (currentWindow.showcaseWindows[0] == parentWindow))) {
          currentWindow.closeWindow();
        }
      } while (windowIter.hasMoreElements());
    }

    // Close embedded
    windowIter = wm.getEnumerator('navigator:browser');
    while (windowIter.hasMoreElements()) {
      var currentWindow = windowIter.getNext();
      if ((currentWindow.embeddedShowcaseWindow) && 
        (
        (currentWindow.embeddedShowcaseWindow.globalMode) ||
        (currentWindow.embeddedShowcaseWindow.showcaseWindows && (currentWindow.embeddedShowcaseWindow.showcaseWindows.length > 0) && currentWindow.embeddedShowcaseWindow.showcaseWindows[0] && (currentWindow.embeddedShowcaseWindow.showcaseWindows[0] == parentWindow))
        )
        ) {
        currentWindow.embeddedShowcaseWindow.closeWindow();
      }
    }
  }

  if (topWindow != window) {
    // We're embedded (tab or sidebar)
    topWindow.embeddedShowcaseWindow = window;
    if (window.top == topWindow) {
      // Sidebar
      var showcaseBox = document.getElementById("showcase-box");
      showcaseBox.style.backgroundColor = tabCustomBackgroundColor + " !important;";
      showcaseBox.setAttribute("focusable", "true");
      showcaseBox.focus();
    }
  }


  loadSettings();

  if ((window.top != topWindow) && tabCustomColors) {
    // We're a tab, set the style...
    var tabbrowser = topWindow.document.getElementById("content");
    var l = tabbrowser.browsers.length;
    for (var i = 0; i < l; i++) {
      var tab = tabbrowser.mTabContainer.childNodes[i];
      if (tab.linkedBrowser.contentWindow == window) {
        setTabColor(tab);
        var showcaseBox = document.getElementById("showcase-box");
        showcaseOldBackgroundColor = showcaseBox.style.backgroundColor;
        showcaseBox.style.backgroundColor = tabCustomBackgroundColor + " !important;";
      }
    }
    window.focus();
  }

  if (windowAttributeTranslucent) {
    try {
      var translucentWindowCss = window.document.createProcessingInstruction("xml-stylesheet","type=\"text/css\" href=\"chrome://showcase/skin/translucentWindow.css\"");
      window.hidden = true;
      window.document.insertBefore(translucentWindowCss, window.document.documentElement);
    } catch (e) { alert("startup - " + e); }
  }

  if ((topWindow == window) && (windowAttributeFullScreen || windowAttributeMaximized)) {
    // Invoke it "later", when window is created. If we don't do this, we won't overlap the taskbar in Windows
    setTimeout(activateFullScreen, 0);
  } else if (closedEmbedded) {
    setTimeout(contentStartup, 0, thumbnailActiveAnimation);
  } else {
    // Proceed
    contentStartup(thumbnailActiveAnimation);
  }
}

// To unlink a tab from the events structure
function unlinkTab(targetTab, tabWindow) {
  try {
  if (targetTab.showcaseCanvasBox) {
    // Remove from current animations if it's there
    if (thumbnailActiveAnimation) {
      var animationPos = currentAnimations.indexOf(targetTab.showcaseCanvasBox);
      if (animationPos > -1) {
        currentAnimations.removeElementAt(animationPos);
      }
    }

    if (currentCursor == targetTab) {
      currentCursor = null;
      currentCursorWindow = null;
    }

    while (targetTab.showcaseCanvasBox.hasChildNodes()) {
      targetTab.showcaseCanvasBox.removeChild(targetTab.showcaseCanvasBox.firstChild);
    }
    targetTab.showcaseCanvasBox.showcaseCanvas = undefined;
    targetTab.showcaseCanvasBox.showcaseTab = undefined;
    targetTab.showcaseCanvasBox.showcaseLabel = undefined;
    targetTab.showcaseCanvasBox.showcaseImageLabel = undefined;
    targetTab.showcaseCanvasBox.showcaseCloseButton = undefined;
    targetTab.showcaseCanvasBox = undefined;

    if (targetTab.showcaseCanvasSubBox) {
      while (targetTab.showcaseCanvasSubBox.hasChildNodes()) {
        targetTab.showcaseCanvasSubBox.removeChild(targetTab.showcaseCanvasSubBox.firstChild);
      }
      targetTab.showcaseCanvasSubBox = undefined;
    }
  }
  
  targetTab.showcaseIETab = undefined;

  if (targetTab.showcaseProgressListener) {
    targetTab.showcaseProgressListener.unloadListener();
    targetTab.showcaseProgressListener = undefined;
  }

  if (targetTab.showcaseTabEventListener) {
    try {
//      targetTab.contentWindow.removeEventListener("resize", targetTab.showcaseTabEventListener, false);
      targetTab.removeEventListener("resize", targetTab.showcaseTabEventListener, false);
      targetTab.removeEventListener("DOMTitleChanged", targetTab.showcaseTabEventListener, false);
    } catch (e) {}
    targetTab.showcaseTabEventListener = undefined;
  }
  
  if (targetTab.showcaseCanvas) {
    targetTab.showcaseCanvas.width = 0;
    targetTab.showcaseCanvas.height = 0;
    targetTab.showcaseCanvas.showcaseCanvasBox = undefined;
    targetTab.showcaseCanvas = undefined;
  }

  if (tabWindow.showcaseTabs) {
    var position = tabWindow.showcaseTabs.indexOf(targetTab);
    if (position > -1) {
      tabWindow.showcaseTabs.removeElementAt(position);
    }
  }
  } catch (e) { alert("unlinkTab: " + e); }
}

function setTabColor(tab) {
  var style = "-moz-appearance: none !important; color: " + tabCustomTitleColor + " !important; background: " + tabCustomBackgroundColor + " !important;";
  tabOldStyle = tab.getAttribute("style");
  tab.setAttribute("style",  tabOldStyle + style);
}


function updateBorders(targetWindow) {
  // Get selected tab
  var tabbrowser = targetWindow.document.getElementById("content");
  var selectedTab = tabbrowser.browsers.item(tabbrowser.tabContainer.selectedIndex);
  
  var multipleWindows = (showcaseWindows.length > 1);

  // Iterate array
  for (var c=0; c<targetWindow.showcaseTabs.length; c++) {
    var targetTab = targetWindow.showcaseTabs[c];
    
    if (targetTab.showcaseCanvasBox) {
      if (multipleWindows && coloredThumbnailsForMultipleWindows) {
        if (targetTab == selectedTab) {
          targetTab.showcaseCanvasSubBox.setAttribute("class", "coloredSelectedThumbnailBox" + targetWindow.showcaseColor);
        } else {
          targetTab.showcaseCanvasSubBox.setAttribute("class", "coloredThumbnailBox" + targetWindow.showcaseColor);
        }
      } else {
        if (targetTab == selectedTab) {
          targetTab.showcaseCanvasSubBox.setAttribute("class", "selectedThumbnailBox");
        } else {
          targetTab.showcaseCanvasSubBox.setAttribute("class", "thumbnailBox");
        }
      }
    }
  }
}

function updateTabBorder(targetTab, targetWindow) {
  // Get selected tab
  var tabbrowser = targetWindow.document.getElementById("content");
  var selectedTab = tabbrowser.browsers.item(tabbrowser.tabContainer.selectedIndex);
  
  var multipleWindows = (showcaseWindows.length > 1);
  if (multipleWindows && coloredThumbnailsForMultipleWindows) {
    if (targetTab == selectedTab) {
      targetTab.showcaseCanvasSubBox.setAttribute("class", "coloredSelectedThumbnailBox" + targetWindow.showcaseColor);
    } else {
      targetTab.showcaseCanvasSubBox.setAttribute("class", "coloredThumbnailBox" + targetWindow.showcaseColor);
    }
  } else {
    if (targetTab == selectedTab) {
      targetTab.showcaseCanvasSubBox.setAttribute("class", "selectedThumbnailBox");
    } else {
      targetTab.showcaseCanvasSubBox.setAttribute("class", "thumbnailBox");
    }
  }

  if (targetTab == currentCursor) {
    targetTab.showcaseCanvasSubBox.style.borderStyle = 'dashed';
  } else {
    targetTab.showcaseCanvasSubBox.style.borderStyle = 'solid';
  }
}

function checkWindow(targetWindow) {
  try {
    if (linkWindow(targetWindow)) {
      setTimeout(refreshContent, 0);
    }
  } catch (e) {
    // Window not ready, try again later...
    setTimeout(checkWindow, 100, targetWindow);
  }
}

function linkWindow(targetWindow) {
  if (!targetWindow.showcaseProcessed) {
      var tabbrowser = targetWindow.document.getElementById("content");
      var showcaseBox = document.getElementById("showcase-box");

    // Assign window color
    targetWindow.showcaseColor = windowCurrentColor;
    windowCurrentColor++;
    if (windowCurrentColor > NUMBER_BORDER_COLORS)
      windowCurrentColor = 1;
      

      // Add listeners to tabsDOMNodeInserted
      var tabsChangeListener = new TabsChangeListener(targetWindow);
      tabbrowser.tabContainer.addEventListener("select", tabsChangeListener, false);
      tabbrowser.tabContainer.addEventListener("DOMNodeInserted", tabsChangeListener, false);
      tabbrowser.tabContainer.addEventListener("DOMNodeRemoved", tabsChangeListener, false); 
      targetWindow.showcaseTabsChangeListener = tabsChangeListener;

      targetWindow.showcaseLastGoodWidth = tabbrowser.browsers.item(tabbrowser.tabContainer.selectedIndex).contentWindow.innerWidth;
      targetWindow.showcaseLastGoodHeight = tabbrowser.browsers.item(tabbrowser.tabContainer.selectedIndex).contentWindow.innerHeight;

      targetWindow.showcaseTabs = new Array();
      // link current tabs
      for (var j = 0; j< tabbrowser.browsers.length; j++) {
        var targetTab = tabbrowser.browsers.item(j);
        if (targetTab.contentWindow != window) {
          if (!targetTab.showcaseCanvasBox) {
//          var newThumbnail = createThumbnail(targetTab, tabbrowser, (j == tabbrowser.tabContainer.selectedIndex), (windows.length > 1), targetWindow.showcaseColor);
            var newThumbnail = createThumbnail(targetTab, tabbrowser, targetWindow);
            showcaseBox.appendChild(newThumbnail);
          }
          targetWindow.showcaseTabs.push(targetTab);
        }
      }

    targetWindow.showcaseProcessed = true;
    showcaseWindows.push(targetWindow);

    return true;
  }
  return false;
}

// To unlink a window
function unlinkWindow(targetWindow) {
  if (targetWindow.showcaseProcessed) {
    var tabbrowser = targetWindow.document.getElementById("content");
    targetWindow.showcaseColor = undefined;

    // Remove listeners to tabsDOMNodeInserted
    tabbrowser.tabContainer.removeEventListener("select", targetWindow.showcaseTabsChangeListener, false); 
    tabbrowser.tabContainer.removeEventListener("DOMNodeInserted", targetWindow.showcaseTabsChangeListener, false); 
    tabbrowser.tabContainer.removeEventListener("DOMNodeRemoved", targetWindow.showcaseTabsChangeListener, false); 
    targetWindow.showcaseTabsChangeListener = undefined;
    
    // Unlink tabs
    var showcaseBox = document.getElementById("showcase-box");
    while (targetWindow.showcaseTabs.length > 0) {
      var targetTab = targetWindow.showcaseTabs[0];
      showcaseBox.removeChild(targetTab.showcaseCanvasBox);
      unlinkTab(targetTab, targetWindow);
    }

    targetWindow.showcaseProcessed = undefined;

    var position = showcaseWindows.indexOf(targetWindow);
    if (position > -1) {
      showcaseWindows.removeElementAt(position);
    }

    return true;
  }

  return false;
}

function unlinkShowcase() {
  // Reset tab style
  if ((window.top != topWindow) && tabCustomColors) {
    var showcaseBox = document.getElementById("showcase-box");
    showcaseBox.style.backgroundColor = showcaseOldBackgroundColor;

    // We're a tab, set the style...
    var tabbrowser = topWindow.document.getElementById("content");
    var l = tabbrowser.browsers.length;
    for (var i = 0; i < l; i++) {
      var tab = tabbrowser.mTabContainer.childNodes[i];
      if (tab.linkedBrowser.contentWindow == window) {
        tab.setAttribute("style", tabOldStyle);
      }
    }
  }

  // Unload all pending variables
  try {
    unregisterObservers();
    while (showcaseWindows.length > 0) {
      unlinkWindow(showcaseWindows[0]);
    }

    currentAnimations.removeAllElements();
    pendingThumbnails.removeAllElements();

    if (topWindow != window) {
      topWindow.embeddedShowcaseWindow = undefined;
    }
  } catch (e) {
    alert("unlinkShowcase - " + e);
  }
  return true; 
}

function registerObservers() {
  // We want to observe click actions in our window
  addEventListener("mouseup", mouseUpWindow, (topWindow != window));
  addEventListener("keydown", keyboardSpaceDown, false);
  addEventListener("keyup", keyboardSpaceUp, false);

  // If we're watching multiple windows...
  if (globalMode) {
    var watcherService =
      Components.classes["@mozilla.org/embedcomp/window-watcher;1"]
        .getService(Components.interfaces.nsIWindowWatcher);
    watcherService.registerNotification(windowObserver);
  }
}

function unregisterObservers() {
  // Unregister mouse listener
  removeEventListener("mouseup", mouseUpWindow, (topWindow != window));
  removeEventListener("keydown", keyboardSpaceDown, false);
  removeEventListener("keyup", keyboardSpaceUp, false);

  // If we're watching multiple windows, unregister ourselves
  if (globalMode) {
    var watcherService =
      Components.classes["@mozilla.org/embedcomp/window-watcher;1"]
        .getService(Components.interfaces.nsIWindowWatcher);
    watcherService.unregisterNotification(windowObserver);
  }
}

function loadSettings() {
  // Load showcase settings
  var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                getService(Components.interfaces.nsIPrefService);
  prefs = prefs.getBranch("extensions.showcase.");
  
  coloredThumbnailsForMultipleWindows = prefs.getBoolPref("coloredThumbnailsForMultipleWindows");
  paintThumbnailsAfterLoading = prefs.getBoolPref("paintThumbnailsAfterLoading");
  paintSelectedTabsFirst = prefs.getBoolPref("paintSelectedTabsFirst");
  thumbnailActiveAnimation = prefs.getBoolPref("thumbnailActiveAnimation");
  animationFrames = prefs.getIntPref("animationFrames");
  thumbnailIdleSizeFactor = parseFloat(prefs.getCharPref("thumbnailIdleSizeFactor"));
  thumbnailActiveSizeFactor = parseFloat(prefs.getCharPref("thumbnailActiveSizeFactor"));
  thumbnailShowLabels = prefs.getBoolPref("thumbnailShowLabels");
  thumbnailLabelStyle = prefs.getIntPref("thumbnailLabelStyle");
  windowAttributeFullScreen = prefs.getBoolPref("windowAttributeFullScreen");
  windowAttributeMaximized = prefs.getBoolPref("windowAttributeMaximized");
  windowAttributeTranslucent = prefs.getBoolPref("windowAttributeTranslucent");
  updateThumbnailWhenContentLoaded = prefs.getBoolPref("updateThumbnailWhenContentLoaded");
  updateThumbnailWhenContentChanges = prefs.getBoolPref("updateThumbnailWhenContentChanges");
  limitContentChangeUpdates = prefs.getBoolPref("limitContentChangeUpdates");
  limitContentChangeUpdatesTime = prefs.getIntPref("limitContentChangeUpdatesTime");
  leftClickAction = prefs.getIntPref("leftClickAction");
  middleClickAction = prefs.getIntPref("middleClickAction");

  if (thumbnailActiveAnimation && (thumbnailIdleSizeFactor > thumbnailActiveSizeFactor)) {
    // if idle size is bigger than active size, then we've a mistake here... deactivate animation
    thumbnailActiveAnimation = false;
  }

  if (window != window.top) {
    // Sidebar
    showcaseCloseWhenBlur = false;
    showcaseCloseWhenSelect = prefs.getBoolPref("sidebarCloseWhenSelect");
    showcaseCloseWhenZeroThumbnails = false;
  } else if (window.top != topWindow) {
    // Tab
    tabCustomColors = prefs.getBoolPref("tabCustomColors");
    tabCustomTitleColor = prefs.getCharPref("tabCustomTitleColor");
    tabCustomBackgroundColor = prefs.getCharPref("tabCustomBackgroundColor");
    showcaseCloseWhenBlur = prefs.getBoolPref("tabCloseWhenBlur");
    showcaseCloseWhenSelect = prefs.getBoolPref("tabCloseWhenSelect");
    showcaseCloseWhenZeroThumbnails = prefs.getBoolPref("tabCloseWhenZeroThumbnails");
  } else {
    // Window
    showcaseCloseWhenBlur = prefs.getBoolPref("windowCloseWhenBlur");
    showcaseCloseWhenSelect = prefs.getBoolPref("windowCloseWhenSelect");
    showcaseCloseWhenZeroThumbnails = prefs.getBoolPref("windowCloseWhenZeroThumbnails");
  }
}

function activateFullScreen() {
  if (windowAttributeMaximized) {
    window.maximize();
  }
  if (windowAttributeFullScreen) {
    window.fullScreen = true;
  }
  contentStartup(thumbnailActiveAnimation);
}

function contentStartup(startAnimationThread) {
  registerObservers();

  createContent();

  window.addEventListener("resize", sizeChanged, false); 

  // Start animation thread
  if (startAnimationThread) {
    setTimeout(animationThread, ANIMATION_INTERVAL);
  }
}

function closeWindow() {
  if (window.top != window) {
    if (!closingShowcase) {
      closingShowcase = true;
      unlinkShowcase();
    }
    // Sidebar
    if (globalMode) {
      window.top.toggleSidebar(SHOWCASE_SIDEBAR);
    } else {
      window.top.toggleSidebar(SHOWCASETHISWINDOW_SIDEBAR);
    }
  } else {
    window.close();
  }
}

function showcaseLostFocus() {
  if (showcaseCloseWhenBlur) {
    closeWindow();
  }
}

function showcaseBrowserStop() {
  try {
    const stopFlags = nsIWebNavigation.STOP_ALL;
    window.showcasePopupTab.webNavigation.stop(stopFlags);
  }
  catch(ex) {
  }
}

function showcaseBrowserReloadSkipCache() {
  showcasePopupTab.webNavigation.reload(RELOAD_SKIP_CACHE_FLAGS);
}

function showcaseBrowserReload() {
  showcasePopupTab.webNavigation.reload(RELOAD_FLAGS);
}

function showcaseAddBookmark() {
  try {
    addBookmarkForBrowser(window.showcasePopupTab.webNavigation, false);
  } catch (e) {
    alert(e);
  }
}

function showcasePrint() {
  try {
    window.showcasePopupTab.contentWindow.print();
  } catch (e) {
    alert(e);
  }
}

function openSelectedTab() {
  openTab(showcasePopupTab.showcaseCanvas);
}

function closeSelectedTab() {
  // Close the selected tab
  showcasePopupTab.contentWindow.close();
}

function closeSelectedWindow() {
  // Look for the window that has the tab
  var found = false;
  var targetWindow;
  for (var c=0; ((c<showcaseWindows.length) && !found); c++) {
    if (showcaseWindows[c].showcaseTabs.contains(showcasePopupTab)) {
      targetWindow = showcaseWindows[c];
      found = true;
    }
  }
  if (found) {
    if (typeof(targetWindow.BrowserTryToCloseWindow) == "function") {
      targetWindow.BrowserTryToCloseWindow();
    } else {
      targetWindow.close();
    }
  }
}

function getFirstCursor() {
  if ((!showcaseWindows) || (showcaseWindows.length < 1) || !showcaseWindows[0] || !showcaseWindows[0].showcaseTabs || (showcaseWindows[0].showcaseTabs.length < 1)) {
    return null;
  }
  return new DualArgument(showcaseWindows[0], showcaseWindows[0].showcaseTabs[0]);
}

function getLastCursor() {
  if ((!showcaseWindows) || (showcaseWindows.length < 1) || !showcaseWindows[showcaseWindows.length - 1] || !showcaseWindows[showcaseWindows.length - 1].showcaseTabs || (showcaseWindows[showcaseWindows.length - 1].showcaseTabs.length < 1)) {
    return null;
  }
  return new DualArgument(showcaseWindows[showcaseWindows.length - 1], showcaseWindows[showcaseWindows.length - 1].showcaseTabs[showcaseWindows[showcaseWindows.length - 1].showcaseTabs.length - 1]);
}

function getTabAt(targetRow, targetColumn) {
  // Iterate...
  var found = false;
  var targetWindow;
  var targetTab;
  for (var c=0; ((c<showcaseWindows.length) && !found); c++) {
    targetWindow = showcaseWindows[c];
    for (var d=0; ((d<targetWindow.showcaseTabs.length) && !found); d++) {
      targetTab = targetWindow.showcaseTabs[d];
      if ((targetTab.showcaseRow == targetRow) && (targetTab.showcaseColumn == targetColumn)) {
        found = true;
      }
    }
    if (showcaseWindows[c].showcaseTabs.contains(showcasePopupTab)) {
      found = true;
    }
  }

  return new DualArgument(targetWindow, targetTab);
}

function keyboardLeft() {
  if (inZoomMode)
    return;

  if (!currentCursor) {
    var result = getLastCursor();
    if (result) {
      currentCursor = result.secondArgument;
      currentCursorWindow = result.firstArgument;
      if (currentCursor) {
        updateTabBorder(result.secondArgument, result.firstArgument);
        animateBox(currentCursor.showcaseCanvasBox, true);
      }
    }
  } else {
    var targetRow = currentCursor.showcaseRow;
    var targetColumn = currentCursor.showcaseColumn;

    if (targetColumn > 0) {
      targetColumn--;
    } else {
      targetColumn = mosaicColumns - 1;
    }
    var result = getTabAt(targetRow, targetColumn);
    var targetWindow = result.firstArgument;
    var targetTab = result.secondArgument;

    if (targetTab && (targetTab != currentCursor)) {
      var oldCursor = currentCursor;
      var oldCursorWindow = currentCursorWindow;
      currentCursor = targetTab;
      currentCursorWindow = targetWindow;
      animateBox(oldCursor.showcaseCanvasBox, false);
      updateTabBorder(oldCursor, oldCursorWindow);
      animateBox(currentCursor.showcaseCanvasBox, true);
      updateTabBorder(currentCursor, currentCursorWindow);
    }
  }
}

function keyboardRight() {
  if (inZoomMode)
    return;

  if (!currentCursor) {
    var result = getFirstCursor();
    if (result) {
      currentCursor = result.secondArgument;
      currentCursorWindow = result.firstArgument;
      if (currentCursor) {
        updateTabBorder(result.secondArgument, result.firstArgument);
        animateBox(currentCursor.showcaseCanvasBox, true);
      }
    }
  } else {
    var targetRow = currentCursor.showcaseRow;
    var targetColumn = currentCursor.showcaseColumn;

    if (targetColumn < mosaicColumns - 1) {
      targetColumn++;
    } else {
      targetColumn = 0;
    }
    var result = getTabAt(targetRow, targetColumn);

    if (result.secondArgument == currentCursor) {
      result = getTabAt(targetRow, 0);
    }

    var targetWindow = result.firstArgument;
    var targetTab = result.secondArgument;

    if (targetTab && (targetTab != currentCursor)) {
      var oldCursor = currentCursor;
      var oldCursorWindow = currentCursorWindow;
      currentCursor = targetTab;
      currentCursorWindow = targetWindow;
      animateBox(oldCursor.showcaseCanvasBox, false);
      updateTabBorder(oldCursor, oldCursorWindow);
      animateBox(currentCursor.showcaseCanvasBox, true);
      updateTabBorder(currentCursor, currentCursorWindow);
    }
  }
}

function keyboardUp() {
  if (inZoomMode)
    return;

  if (!currentCursor) {
    var result = getLastCursor();
    if (result) {
      currentCursor = result.secondArgument;
      currentCursorWindow = result.firstArgument;
      if (currentCursor) {
        updateTabBorder(result.secondArgument, result.firstArgument);
        animateBox(currentCursor.showcaseCanvasBox, true);
      }
    }
  } else {
    var targetRow = currentCursor.showcaseRow;
    var targetColumn = currentCursor.showcaseColumn;

    if (targetRow > 0) {
      targetRow--;
    } else {
      targetRow = mosaicRows - 1;
    }
    var result = getTabAt(targetRow, targetColumn);

    if ((result.secondArgument) && (result.secondArgument.showcaseColumn != currentCursor.showcaseColumn)) {
      if (mosaicRows > 1) {
        result = getTabAt(mosaicRows - 2, targetColumn);
      } else {
        result = getTabAt(mosaicRows - 1, targetColumn);
      }
    }

    var targetWindow = result.firstArgument;
    var targetTab = result.secondArgument;

    if ((targetTab) && (targetTab != currentCursor)) {
      var oldCursor = currentCursor;
      var oldCursorWindow = currentCursorWindow;
      currentCursor = targetTab;
      currentCursorWindow = targetWindow;
      animateBox(oldCursor.showcaseCanvasBox, false);
      updateTabBorder(oldCursor, oldCursorWindow);
      animateBox(currentCursor.showcaseCanvasBox, true);
      updateTabBorder(currentCursor, currentCursorWindow);
    }
  }
}

function keyboardDown() {
  if (inZoomMode)
    return;

  if (!currentCursor) {
    var result = getFirstCursor();
    if (result) {
      currentCursor = result.secondArgument;
      currentCursorWindow = result.firstArgument;
      if (currentCursor) {
        updateTabBorder(result.secondArgument, result.firstArgument);
        animateBox(currentCursor.showcaseCanvasBox, true);
      }
    }
  } else {
    var targetRow = currentCursor.showcaseRow;
    var targetColumn = currentCursor.showcaseColumn;

    if (targetRow < mosaicRows - 1) {
      targetRow++;
    } else {
      targetRow = 0;
    }
    var result = getTabAt(targetRow, targetColumn);

    if ((result.secondArgument) && (result.secondArgument.showcaseColumn != currentCursor.showcaseColumn)) {
      result = getTabAt(0, targetColumn);
    }

    var targetWindow = result.firstArgument;
    var targetTab = result.secondArgument;

    if (targetTab && (targetTab != currentCursor)) {
      var oldCursor = currentCursor;
      var oldCursorWindow = currentCursorWindow;
      currentCursor = targetTab;
      currentCursorWindow = targetWindow;
      animateBox(oldCursor.showcaseCanvasBox, false);
      updateTabBorder(oldCursor, oldCursorWindow);
      animateBox(currentCursor.showcaseCanvasBox, true);
      updateTabBorder(currentCursor, currentCursorWindow);
    }
  }
}

function keyboardEnter() {
  if (!currentCursor)
    return;
  openTab(currentCursor.showcaseCanvasBox.showcaseCanvas);
}

function keyboardSpaceDown(ev) {
  if (ev.keyCode == VK_SPACE) {
    if (!currentCursor)
      return;
  
    enterZoomMode(currentCursor);
  }
}

function keyboardSpaceUp(ev) {
  if (ev.keyCode == VK_SPACE) {
    if (!currentCursor)
      return;
  
    leaveZoomMode();
  }
}

function keyboardDelete() {
  if (!currentCursor)
    return;

  if (inZoomMode)
    leaveZoomMode();
  
  currentCursor.contentWindow.close();
}

function toggleFullScreen() {
  if (topWindow == window) {
    window.fullScreen = !window.fullScreen;
  }
}

function createHTMLElement(aElement, aId) {
  var item = document.createElementNS(HTML_NS, aElement);
  item.setAttribute("id", aId);
  return item;
}

function createXULElement(aElement, aId) {
  var item = document.createElementNS(XUL_NS, aElement);
  item.setAttribute("id", aId);
  return item;
}

function setSize(targetElement,width,height) {
  targetElement.style.width = width+"px";
  targetElement.style.height = height+"px";
//  targetElement.style.minWidth = targetElement.style.width;
//  targetElement.style.minHeight = targetElement.style.height;
  targetElement.style.maxWidth = targetElement.style.width;
  targetElement.style.maxHeight = targetElement.style.height;
  targetElement.setAttribute("width", width);
  targetElement.setAttribute("height", height);
}

function createContent() {
  try {
    var windows = getTargetWindows();
    var shouldRefresh = false;

    for (var i = 0; i < windows.length; i++) {
      try {
        if (linkWindow(windows[i])) {
          shouldRefresh = true;
        }
      } catch (e) {
        // We failed, check window later
        setTimeout(checkWindow, 100, windows[i]);
      }
    }

    for (var j=0; j<showcaseWindows.length; j++) {
      updateBorders(showcaseWindows[j]);
    }

    if (shouldRefresh) {
      updateTitle();
//      refreshContent(windows);
      setTimeout(refreshContent, 0);
    }
  } catch (e) {
      alert("createContent - " + e);
  }
}

function delayedCreateContent() {
  setTimeout("createContent();", 0);
}

function createThumbnail(targetTab, tabbrowser, tabwindow) {
  var newId = idGenerator++;

  var canvas = createHTMLElement("canvas", "canvas-" + newId);
  canvas.title = " ";
  canvas.style.MozBoxFlex = "1";
  canvas.setAttribute("class", "thumbnail");
  var canvasBox;
  var subCanvasBox;

  if (thumbnailShowLabels) {
    if (thumbnailLabelStyle == 0) {
      canvasBox = createXULElement("groupbox", "box-" + newId);
      var canvasBoxLabel = createXULElement("caption", "boxcaption-" + newId);
      canvasBoxLabel.style.MozBoxFlex = "1";
      canvasBoxLabel.style.MozBoxAlign = "center";
      canvasBoxLabel.style.MozBoxPack = "center";
      canvasBox.showcaseImageLabel = createXULElement("image", "boximage-" + newId);
      if (usingSeaMonkey) {
        canvasBox.showcaseImageLabel.setAttribute("class", "sea-thumbnail-icon");
      } else {
        canvasBox.showcaseImageLabel.setAttribute("class", "thumbnail-icon");
      }
      canvasBox.showcaseImageLabel.setAttribute("validate", "never");
      canvasBox.showcaseImageLabel.setAttribute("onerror", "this.src=null");
      canvasBox.showcaseImageLabel.style.MozBoxFlex = "0";
      canvasBox.showcaseLabel = createXULElement("label", "boxlabel-" + newId);
      canvasBox.showcaseLabel.setAttribute("crop", "end");
      canvasBox.showcaseLabel.style.MozBoxFlex = "1";
      canvasBox.showcaseLabelSpacer = createXULElement("caption", "boxlabelspacer-" + newId);
      canvasBox.showcaseLabelSpacer.style.MozBoxFlex = "10000";
      subCanvasBox = createXULElement("box", "subbox-" + newId);
      subCanvasBox.style.MozBoxFlex = "1";
      canvasBoxLabel.appendChild(canvasBox.showcaseImageLabel);
      canvasBoxLabel.appendChild(canvasBox.showcaseLabel);
      canvasBox.appendChild(canvasBoxLabel);
      canvasBox.appendChild(canvasBox.showcaseLabelSpacer);
      canvasBox.appendChild(subCanvasBox);
      subCanvasBox.appendChild(canvas);
    } else if ((thumbnailLabelStyle == 1) || (thumbnailLabelStyle == 2)) {
      canvasBox = createXULElement("vbox", "box-" + newId);
      canvasBox.MozBoxAlign = "stretch";
      var canvasBoxLabel = createXULElement("hbox", "boxcaption-" + newId);
      canvasBoxLabel.setAttribute("class", "thumbnail-label");
      canvasBoxLabel.style.MozBoxFlex = "1";
      canvasBoxLabel.style.MozBoxAlign = "center";
      canvasBoxLabel.style.MozBoxPack = "center";
      canvasBox.showcaseImageLabel = createXULElement("image", "boximage-" + newId);
      if (usingSeaMonkey) {
        canvasBox.showcaseImageLabel.setAttribute("class", "sea-thumbnail-icon");
      } else {
        canvasBox.showcaseImageLabel.setAttribute("class", "thumbnail-icon");
      }
      canvasBox.showcaseImageLabel.setAttribute("validate", "never");
      canvasBox.showcaseImageLabel.setAttribute("onerror", "this.src=null");
      //canvasBox.showcaseImageLabel.style.MozBoxFlex = "0";
      canvasBox.showcaseLabel = createXULElement("label", "boxlabel-" + newId);
      canvasBox.showcaseLabel.setAttribute("crop", "end");
      canvasBox.showcaseLabel.style.MozBoxFlex = "1";
      //canvasBox.showcaseLabel.crop = "end";
      var showcaseLabelSpacer = createXULElement("separator", "boxlabelspacer-" + newId);
      showcaseLabelSpacer.style.MozBoxFlex = "10000";
      showcaseLabelSpacer.setAttribute("class", "groove");

      var canvasBoxLabelRecipient = createXULElement("caption", "labelrecipient-" + newId);
      canvasBoxLabelRecipient.style.MozBoxFlex = "1";
      canvasBoxLabelRecipient.style.marginLeft = "0";
      canvasBoxLabelRecipient.style.MozBoxAlign = "center";
      canvasBoxLabelRecipient.style.MozBoxPack = "center";
      canvasBoxLabelRecipient.appendChild(canvasBox.showcaseImageLabel);
      canvasBoxLabelRecipient.appendChild(canvasBox.showcaseLabel);

      subCanvasBox = createXULElement("box", "subbox-" + newId);
      subCanvasBox.style.MozBoxFlex = "1";
      canvasBoxLabel.appendChild(canvasBoxLabelRecipient);
      canvasBoxLabel.appendChild(showcaseLabelSpacer);

      if (thumbnailLabelStyle == 2) {
        var showcaseCloseButton = createXULElement("image", "boxclose-" + newId);
        showcaseCloseButton.targetTab = targetTab;
        showcaseCloseButton.style.MozBoxFlex = "0";
        if (usingSeaMonkey) {
          showcaseCloseButton.setAttribute("class", "sea-thumbnail-close");
        } else {
          showcaseCloseButton.setAttribute("class", "thumbnail-close");
        }

        if (!closeTabTitle) {
          closeTabTitle = document.getElementById("popupClose").getAttribute("label");
        }
        showcaseCloseButton.setAttribute("tooltiptext", closeTabTitle);
        showcaseCloseButton.addEventListener("click", mouseClickClose, false);
        canvasBoxLabel.appendChild(showcaseCloseButton);
        canvasBox.showcaseCloseButton = showcaseCloseButton;
      }

      canvasBox.appendChild(canvasBoxLabel);
      canvasBox.appendChild(subCanvasBox);
      subCanvasBox.appendChild(canvas);
    }
  } else {
    canvasBox = createXULElement("box", "box-" + newId);
    canvasBox.appendChild(canvas);
    subCanvasBox = canvasBox;
  }
  canvasBox.style.MozBoxFlex = "0";

  if (thumbnailActiveAnimation) {
    canvasBox.addEventListener("mouseover", mouseOverCanvas, false);
    canvasBox.addEventListener("mouseout", mouseOutCanvas, false);
  }
  canvasBox.setAttribute("context", "tabMenu");
  canvas.addEventListener("click", mouseClickCanvas, (topWindow != window));
  canvas.addEventListener("mousedown", mouseDownCanvas, (topWindow != window));
  canvas.addEventListener("mouseup", mouseUpCanvas, (topWindow != window));
  canvas.MozBoxFlex = "1";

  targetTab.showcaseCanvas = canvas;
  targetTab.showcaseCanvas.showcaseCanvasBox = canvasBox;
  targetTab.showcaseCanvasBox = canvasBox;
  targetTab.showcaseCanvasBox.showcaseCanvas = canvas;
  targetTab.showcaseCanvasBox.showcaseTab = targetTab;

  targetTab.showcaseCanvasSubBox = subCanvasBox;

  targetTab.showcaseProgressListener = new TabProgressListener(targetTab);
  var targetTabContainerTab = tabbrowser.ownerDocument.getAnonymousElementByAttribute(tabbrowser,"linkedpanel", targetTab.parentNode.id);
  if (targetTabContainerTab && targetTabContainerTab.hasAttribute("busy")) {
    targetTab.showcaseProgressListener.progressListener.isLoading = true;
    updateLoadingState(targetTab);
  }

  if ((targetTab.contentDocument) && (targetTab.contentDocument.URL)) {
    if (targetTab.contentDocument.URL.substr(0, IETAB_URL.length) == IETAB_URL) {
      targetTab.showcaseIETab = true;
      updateIEState(targetTab);
    }
  }

  targetTab.showcaseTabEventListener = new TabEventListener(targetTab, tabwindow);
//  targetTab.contentWindow.addEventListener("resize", targetTab.showcaseTabEventListener, false);
  targetTab.addEventListener("resize", targetTab.showcaseTabEventListener, false);
  targetTab.addEventListener("DOMTitleChanged", targetTab.showcaseTabEventListener, false);
//  targetTab.contentWindow.addEventListener("select", targetTab.showcaseTabEventListener, false);

  return canvasBox;
}

function refreshScreen() {
  if (closingShowcase) {
    return;
  }

  // Unlink showcase
  unlinkShowcase();
  
  // Reset window color
  windowCurrentColor = 1;

  // Startup...
  contentStartup(false);
}

function refreshContent() {
  var showcaseBox = document.getElementById("showcase-box");

  // Extend showcase box
  if ((window.innerWidth > 0) && (window.innerHeight > 0)) {
    setSize(showcaseBox, window.innerWidth, window.innerHeight);
  }

  // Mosaic code
  mosaicWidth = showcaseBox.width;
  mosaicHeight = showcaseBox.height;

  if (!inZoomMode) {
    var mosaicSizeResult = mosaicCalculateBestGrid(mosaicWidth, mosaicHeight);
    mosaicRows = mosaicSizeResult[0];
    mosaicColumns = mosaicSizeResult[1];
    thumbWidth =  mosaicWidth / mosaicColumns;
    thumbHeight = mosaicHeight / mosaicRows;
  } else {
    thumbWidth =  mosaicWidth;
    thumbHeight = mosaicHeight;
  }

  updateThumbnails(thumbWidth, thumbHeight, thumbnailIdleSizeFactor / 100, thumbnailActiveSizeFactor / 100);
  mosaicRefresh();
}

function sizeChanged(el) {
  refreshContent();
}

function hiddenTabPopup(ep) {
  /*
  if (!window.arguments) {
    // When we close the popup, close the window.
    // If not, there's a bug that causes the mouse click not to be recognized. 
    setTimeout(closeWindow, 0);
  }
  */
}

function showingTabPopup(ep) {
  var popupOpener = document.popupNode;
  if (popupOpener.showcaseCanvasBox) {
    showcasePopupTab = popupOpener.showcaseCanvasBox.showcaseTab;
  } else {
    showcasePopupTab = popupOpener.showcaseTab;
  }

  var found = false;
  var targetWindow;
  for (var c=0; ((c<showcaseWindows.length) && !found); c++) {
    if (showcaseWindows[c].showcaseTabs.contains(showcasePopupTab)) {
      targetWindow = showcaseWindows[c];
      found = true;
    }
  }

  if (found) {
    // Modify popup options depending of item
    var tabbrowser = targetWindow.document.getElementById("content");
    if (tabbrowser.browsers.length > 1) {
      // Hide popupclose, show popupclosetab and popupclosewindow
      document.getElementById("popupClose").setAttribute("hidden", true);
      document.getElementById("popupCloseTab").setAttribute("hidden", false);
      document.getElementById("popupCloseWindow").setAttribute("hidden", false);
    } else {
      // Hide popupclosetab and popupclosewindow, show popupclose
      document.getElementById("popupClose").setAttribute("hidden", false);
      document.getElementById("popupCloseTab").setAttribute("hidden", true);
      document.getElementById("popupCloseWindow").setAttribute("hidden", true);
    }
  } else {
    // Show simple close option
      document.getElementById("popupClose").setAttribute("hidden", false);
      document.getElementById("popupCloseTab").setAttribute("hidden", true);
      document.getElementById("popupCloseWindow").setAttribute("hidden", true);
  }

  var backCommand = document.getElementById("Browser:Back");
  var forwardCommand = document.getElementById("Browser:Forward");
  var reloadCommand = document.getElementById("Browser:Reload");
  var stopCommand = document.getElementById("Browser:Stop");
  var backDisabled = backCommand.hasAttribute("disabled");
  var forwardDisabled = forwardCommand.hasAttribute("disabled");

  if (forwardDisabled == showcasePopupTab.canGoForward) {
    if (forwardDisabled)
      forwardCommand.removeAttribute("disabled");
    else
      forwardCommand.setAttribute("disabled", true);
  }

  if (backDisabled == showcasePopupTab.canGoBack) {
    if (backDisabled)
      backCommand.removeAttribute("disabled");
    else
      backCommand.setAttribute("disabled", true);
  }
  
  var location = showcasePopupTab.contentDocument.URL;
  if (location == "about:blank" || location == "") {
    reloadCommand.setAttribute("disabled", true);
  } else {
    reloadCommand.removeAttribute("disabled");
  }

  if (showcasePopupTab.showcaseProgressListener.progressListener.isLoading) {
    stopCommand.removeAttribute("disabled");
  } else {
    stopCommand.setAttribute("disabled", true);
  }
}

function enterZoomMode(targetTab) {
  if (inZoomMode) {
    return;
  }

  inZoomMode = true;
  zoomTargetThumbnail = targetTab;
  zoomTargetPreviousCursorStyle = zoomTargetThumbnail.showcaseCanvas.style.cursor;
  zoomTargetThumbnail.showcaseCanvas.style.cursor = "-moz-zoom-in";

  // Make other thumbnails invisible
  for (var i = 0; i < showcaseWindows.length; i++) {
    var targetWindow =  showcaseWindows[i];
    for (var j = 0; j<targetWindow.showcaseTabs.length; j++) {
      var currentTab = targetWindow.showcaseTabs[j];
      if (currentTab != zoomTargetThumbnail) {
        currentTab .showcaseCanvasBox.hidden = true;
      }
    }
  }
  
  refreshContent();
}

function leaveZoomMode() {
  if (!inZoomMode) {
    return;
  }

  inZoomMode = false;
  zoomTargetThumbnail.showcaseCanvas.style.cursor = zoomTargetPreviousCursorStyle;
  zoomTargetPreviousCursorStyle = null;
  zoomTargetThumbnail = null;

  // Make other thumbnails visible
  for (var i = 0; i < showcaseWindows.length; i++) {
    var targetWindow =  showcaseWindows[i];
    for (var j = 0; j<targetWindow.showcaseTabs.length; j++) {
      var currentTab = targetWindow.showcaseTabs[j];
      if (currentTab != zoomTargetThumbnail) {
        currentTab.showcaseCanvasBox.hidden = false;
      }
    }
  }

  refreshContent();
}

function mouseOverCanvas(el) {
  var hasEntered = true;
  var currentNode = el.relatedTarget;

  if (currentNode) {
    while (hasEntered && currentNode.parentNode) {
      hasEntered = (el.currentTarget != currentNode.parentNode);
      currentNode = currentNode.parentNode;
    }
  }

  if (hasEntered) {
    animateBox(el.currentTarget, true);
  }
}

function mouseOutCanvas(el) {
  var hasExited = true;
  var currentNode = el.relatedTarget;

  if (currentNode) {
    while (hasExited && currentNode.parentNode) {
      hasExited = (el.currentTarget != currentNode.parentNode);
      currentNode = currentNode.parentNode;
    }
  }

  if (hasExited) {
    animateBox(el.currentTarget, false);
  }
}

function mouseDownCanvas(el) {
  if ((el.button == 1) || (el.button == 0)) {
    var clickAction;
    
    if (el.button == 0) {
      clickAction = leftClickAction;
    } else {
      clickAction = middleClickAction;
    }

    if (clickAction == 1) { 
      // Zoom mode
      var targetTab;
      if (el.target.showcaseCanvas) {
        targetTab = el.target.showcaseCanvas.showcaseCanvasBox.showcaseTab;
      } else {
        targetTab = el.target.showcaseCanvasBox.showcaseTab;
      }
      enterZoomMode(targetTab);

      if (topWindow != window) {
        el.stopPropagation();
      }
    }
  }
}

function mouseUpWindow(el) {
  if ((el.button == 1) || (el.button == 0)) {
    var clickAction;
    if (el.button == 0) {
      clickAction = leftClickAction;
    } else {
      clickAction = middleClickAction;
    }
    if (clickAction == 1) { 
      // Zoom mode
      leaveZoomMode();
      if (topWindow != window) {
        el.stopPropagation();
      }
    }
  }
}

function mouseUpCanvas(el) {
  if ((el.button == 1) || (el.button == 0)) {
    var clickAction;
    var targetTab;
    if (el.target.showcaseCanvas) {
      targetTab = el.target.showcaseCanvas.showcaseCanvasBox.showcaseTab;
    } else {
      targetTab = el.target.showcaseCanvasBox.showcaseTab;
    }
    
    if (el.button == 0) {
      clickAction = leftClickAction;
    } else {
      clickAction = middleClickAction;
    }


    if (clickAction == 1) { 
      // Zoom mode
      leaveZoomMode();
      if (topWindow != window) {
        el.stopPropagation();
      }
    } else if (el.button == 1) {
      switch (clickAction) { 
        case 0:
          // Close
          targetTab.contentWindow.close();
          if (topWindow != window) {
            el.stopPropagation();
          }
          break;
        case 2:
          // Select
          if (el.target.showcaseCanvas) {
            openTab(el.target.showcaseCanvas);
          } else {
            openTab(el.target);
          }
          if (topWindow != window) {
            el.stopPropagation();
          }
          break;
        case 3:
          // Nothing
          if (topWindow != window) {
            el.stopPropagation();
          }
          break;
      }
    }
  }
}

function mouseClickClose(el) {
  if (el.button == 0) {
    if ((el.target.targetTab) && (el.target.targetTab.contentWindow)) {
      el.target.targetTab.contentWindow.close();
    }
  }
}

function mouseClickCanvas(el) {
  if (el.button == 0) {
    var targetTab;
    if (el.target.showcaseCanvas) {
      targetTab = el.target.showcaseCanvas.showcaseCanvasBox.showcaseTab;
    } else {
      targetTab = el.target.showcaseCanvasBox.showcaseTab;
    }

    var clickAction;
    
    if (el.button == 0) {
      clickAction = leftClickAction;
    } else {
      clickAction = middleClickAction;
    }

    switch (clickAction) { 
      case 0:
        // Close
        targetTab.contentWindow.close();
        if (topWindow != window) {
          el.stopPropagation();
        }
        break;
      case 1:
        // Zoom
        if (topWindow != window) {
          el.stopPropagation();
        }
        break;
      case 2:
        // Select
        if (el.target.showcaseCanvas) {
          openTab(el.target.showcaseCanvas);
        } else {
          openTab(el.target);
        }
        if (topWindow != window) {
          el.stopPropagation();
        }
        break;
      case 3:
        // Nothing
        if (topWindow != window) {
          el.stopPropagation();
        }
        break;
    }
  }
}

function openTab(targetCanvas) {
  // Look for the tab that has the canvas
  try {
    var found = false;
    var targetWindow;
    var targetTabIndex;
    var targetTabBrowser;

    for (var i = 0; (i < showcaseWindows.length) && !found; i++) {
      var targetWindow = showcaseWindows[i];
      var tabbrowser = targetWindow.document.getElementById("content");

      for (var j = 0; (j<tabbrowser.browsers.length) && !found; j++) {
        var targetTab = tabbrowser.browsers.item(j);
        if ((targetTab) && (targetTab.showcaseCanvas == targetCanvas)) {
          found = true;
          targetTabIndex = j;
          targetTabBrowser = tabbrowser;
          targetWindow = targetWindow;
        }
      }
    }

    if (found) {
      targetTabBrowser.tabContainer.selectedIndex = targetTabIndex;

      if (showcaseCloseWhenSelect) {
        targetFocusWindow = targetWindow;
      } else {
        try {
          targetWindow.focus();
        } catch (focusEx) {
          showFocusAlert(targetWindow);
        }
      }
    } else {
      // Alert user we haven't found the page
      // TODO: localized string
      alert("The selected browser couldn't be found.");
    }

    // Last, close ourselves
    if (showcaseCloseWhenSelect) {
      closeWindow();
    }

  } catch (e) {
    alert("openTab: " + e);
  }
}

function selectTab(targetTabBrowser, targetTabIndex) {
  targetTabBrowser.tabContainer.selectedIndex = targetTabIndex;
}

function showFocusAlert(targetWindow) {
  try {
    var showcaseBundle = document.getElementById("bundle_showcase");
    targetWindow.alert(showcaseBundle.getString("showcase.focus.alert"));
  } catch (e) {
    // Fail silently
  }
}

function unloadShowcase() {
  if (window.opener) {
    try {
      // Store window values to remember later
      var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                getService(Components.interfaces.nsIPrefBranch);
      prefs.setIntPref("extensions.showcase.windowLastX", window.screenX);
      prefs.setIntPref("extensions.showcase.windowLastY", window.screenY);
      if (window.innerWidth > 0) {
        prefs.setIntPref("extensions.showcase.windowLastWidth", window.innerWidth);
      }
      if (window.innerHeight > 0) {
        prefs.setIntPref("extensions.showcase.windowLastHeight", window.innerHeight);
      }
      var lastRelativeX = window.screenX - window.opener.screenX;
      var lastRelativeY = window.screenY - window.opener.screenY;
      prefs.setIntPref("extensions.showcase.windowLastRelativeX", lastRelativeX);
      prefs.setIntPref("extensions.showcase.windowLastRelativeY", lastRelativeY);
    } catch (ex) {
      // We may have trouble in the case window is closed, so just fail silently
    }
  }

  // Free resources
  if (!closingShowcase) {
    closingShowcase = true;
    unlinkShowcase();

    if (targetFocusWindow) {
      try {
        targetFocusWindow.focus();
      } catch (e) {
        showFocusAlert(targetFocusWindow);
      }
    }
  }
}

function animateBox(targetElement, growBox) {
  var targetBox;

  if (targetElement.showcaseCanvas) { 
    targetBox = targetElement.showcaseCanvas.showcaseCanvasBox;
  } else {
    targetBox = targetElement.showcaseCanvasBox;
  }

  if ((targetBox.showcaseTab == currentCursor) && !growBox)
    return;

  try {
  targetBox.showcaseGrowing = growBox;
  if (!currentAnimations.contains(targetBox)) {
    if (growBox) {
      if (thumbnailActiveSizeFactor > 100) {
        // Hack to bring it to the front
        var showcaseBox = document.getElementById("showcase-box");
        if (showcaseBox.lastChild != targetBox) {
          if (thumbnailActiveAnimation) {
            targetBox.removeEventListener("mouseover", mouseOverCanvas, false);
            targetBox.removeEventListener("mouseout", mouseOutCanvas, false);
            if (targetBox.showcaseCloseButton) {
              targetBox.showcaseCloseButton.removeEventListener("click", mouseClickClose, false);
            }
          }
          showcaseBox.removeChild(targetBox);
          showcaseBox.appendChild(targetBox);
          if (thumbnailActiveAnimation) {
            targetBox.addEventListener("mouseover", mouseOverCanvas, false);
            targetBox.addEventListener("mouseout", mouseOutCanvas, false);
            if (targetBox.showcaseCloseButton) {
              targetBox.showcaseCloseButton.addEventListener("click", mouseClickClose, false);
            }
          }
        }
      }
      targetBox.showcaseAnimationFrame = 0;
    } else {
      targetBox.showcaseAnimationFrame = animationFrames;
    }
    currentAnimations.push(targetBox);
  }
  } catch (e) {
    alert("animateBox - " + e);
  }
}

function animationThread() {
  if (currentAnimations && (currentAnimations.length > 0)) {
    var showcaseBox = document.getElementById("showcase-box");
    var c = 0;

    while (c < currentAnimations.length) {
      var currentBox = currentAnimations[c];
      var endAnimation = false;

      if (currentBox.showcaseGrowing) {
        currentBox.showcaseAnimationFrame++;
        if (currentBox.showcaseAnimationFrame >= animationFrames) {
          currentBox.showcaseAnimationFrame = animationFrames;
          endAnimation = true;
        }
      } else {
        currentBox.showcaseAnimationFrame--;
        if (currentBox.showcaseAnimationFrame <= 0) {
          currentBox.showcaseAnimationFrame = 0;
          endAnimation = true;
        }
      }
      
      var newWidth = currentBox.showcaseIdleWidth + (currentBox.showcaseActiveWidth - currentBox.showcaseIdleWidth) * currentBox.showcaseAnimationFrame / animationFrames;
      var newHeight = currentBox.showcaseIdleHeight + (currentBox.showcaseActiveHeight - currentBox.showcaseIdleHeight) * currentBox.showcaseAnimationFrame / animationFrames;
      
      if (newWidth > showcaseBox.width) {
        // We're getting out of the limits!! Set maximum size that will fit the window...
        newHeight = newHeight * (showcaseBox.width / newWidth);
        newWidth = showcaseBox.width;
      }

      if (newHeight > showcaseBox.height) {
        // We're getting out of the limits!! Set maximum size that will fit the window...
        newWidth = newWidth * (showcaseBox.height / newHeight);
        newHeight = showcaseBox.height;
      }

      var newTop = currentBox.showcaseCenterY - newHeight / 2;
      var newLeft = currentBox.showcaseCenterX - newWidth / 2;
      var limitX = showcaseBox.width - newWidth;
      var limitY = showcaseBox.height - newHeight;

      if (newTop < 0) {
        if ((currentBox.getAttribute("top") == 0) && !currentBox.showcaseGrowing) {
          newTop = -1;
        } else {
          newTop = 0;
        }
      } else if (newTop > limitY) {
        if ((currentBox.getAttribute("top") == limitY) && !currentBox.showcaseGrowing) {
          newTop = limitY + 1;
        } else {
          newTop = limitY;
        }
      }
      if (newLeft < 0) {
        if ((currentBox.getAttribute("left") == 0) && !currentBox.showcaseGrowing) {
          newLeft = -1;
        } else {
          newLeft = 0;
        }
      } else if (newLeft > limitX) {
        if ((currentBox.getAttribute("left") == limitX) && !currentBox.showcaseGrowing) {
          newLeft = limitX + 1;
        } else {
          newLeft = limitX;
        }
      }

      currentBox.setAttribute("top", newTop);
      currentBox.setAttribute("left", newLeft);
      setSize(currentBox, newWidth, newHeight);

      if (endAnimation) {
        currentAnimations.removeElementAt(c);
        // Repaint thumbnail if needed
        if (currentBox.showcaseGrowing) {
          if (currentBox.showcaseCanvas.showcasePainted) {
            var showcaseCanvasScreenBox = document.getBoxObjectFor(currentBox.showcaseCanvas);
            if ((showcaseCanvasScreenBox.width > currentBox.showcaseCanvas.width) || (showcaseCanvasScreenBox.height > currentBox.showcaseCanvas.height)) {
              setTimeout(paintThumbnail, 0, currentBox.showcaseTab);
            }
          }
        }
      } else {
        c++;
      }
    }
  }
  if (!closingShowcase) {
    setTimeout("animationThread()", ANIMATION_INTERVAL);
  }
}

function updateThumbnail(targetThumbnail, thumbWidth, thumbHeight, idleFactor, activeFactor) {
try {
  var canvasActiveW, canvasActiveH;
  var canvasIdleW, canvasIdleH;
  var windowW, windowH;

    windowW = targetThumbnail.contentWindow.innerWidth;
    windowH = targetThumbnail.contentWindow.innerHeight;

    if ((thumbWidth * windowH / windowW) > thumbHeight) {
      canvasIdleW = Math.floor(thumbHeight * windowW * idleFactor / windowH);
      canvasIdleH = Math.floor(thumbHeight * idleFactor);
      canvasActiveW = Math.floor(thumbHeight * windowW * activeFactor / windowH);
      canvasActiveH = Math.floor(thumbHeight * activeFactor);
    } else {
      canvasIdleW = Math.floor(thumbWidth * idleFactor);
      canvasIdleH = Math.floor(thumbWidth * windowH * idleFactor / windowW);
      canvasActiveW = Math.floor(thumbWidth * activeFactor);
      canvasActiveH = Math.floor(thumbWidth * windowH * activeFactor / windowW);
    }

      targetThumbnail.showcaseCanvasBox.showcaseIdleWidth = canvasIdleW;
      targetThumbnail.showcaseCanvasBox.showcaseIdleHeight = canvasIdleH;
      targetThumbnail.showcaseCanvasBox.showcaseActiveWidth = canvasActiveW;
      targetThumbnail.showcaseCanvasBox.showcaseActiveHeight = canvasActiveH;
      setSize(targetThumbnail.showcaseCanvasBox, 
        canvasIdleW, 
        canvasIdleH);
} catch (e) {
  // Probably we failed because the thumbnail has disappeared, so fail silently
}
}

function updateThumbnails(thumbWidth, thumbHeight, idleFactor, activeFactor) {
  var targetItem;

  var pendingSelectedThumbnailIndex = 0;

  if (inZoomMode) {
    targetItem = zoomTargetThumbnail;
    if ((targetItem) && (targetItem.showcaseCanvas) && (targetItem.showcaseCanvasBox)) {
      try {
        updateThumbnail(targetItem, thumbWidth, thumbHeight, 1, 1);
      } catch (e) {
        alert("updateThumbs1: " + e);
      }
      var showcaseCanvasScreenBox = document.getBoxObjectFor(targetItem.showcaseCanvas);
      if (!targetItem.showcaseCanvas.showcasePainted || (showcaseCanvasScreenBox.width > targetItem.showcaseCanvas.width) || (showcaseCanvasScreenBox.height > targetItem.showcaseCanvas.height)) {
        try {
          createTooltip(targetItem);
        } catch (e) {
        // We can fail while creating the tooltip, since window 
        }
        // We're gonna paint it, so set it to false
        targetItem.showcaseCanvas.showcasePainted = false;

        if (paintThumbnailsAfterLoading) {
          // If it's a window's selected thumbnail, give it priority over the others
          if (!pendingThumbnails.contains(targetItem)) {
            pendingThumbnails.insertElementAt(targetItem, 0);
          }
        } else {
          setTimeout(paintThumbnail, 0, targetItem);
        }
      }
    }
  } else {
    for (var i = 0; i < showcaseWindows.length; i++) {
      var targetWindow = showcaseWindows[i];
      if (targetWindow.showcaseTabs) {
        var tabbrowser = targetWindow.document.getElementById("content");
        var currentSelectedTab = tabbrowser.browsers.item(tabbrowser.tabContainer.selectedIndex);

        for (var j = 0; j< targetWindow.showcaseTabs.length; j++) {
          // Resize canvas and canvasBox
          targetItem = targetWindow.showcaseTabs[j];
        
          if ((targetItem) && (targetItem.showcaseCanvas) && (targetItem.showcaseCanvasBox)) {
            try {
              updateThumbnail(targetItem, thumbWidth, thumbHeight, idleFactor, activeFactor);
            } catch (e) {
              alert("updateThumbs1: " + e);
            }

            var showcaseCanvasScreenBox = document.getBoxObjectFor(targetItem.showcaseCanvas);
            if (!targetItem.showcaseCanvas.showcasePainted || (showcaseCanvasScreenBox.width > targetItem.showcaseCanvas.width) || (showcaseCanvasScreenBox.height > targetItem.showcaseCanvas.height)) {
              try {
                createTooltip(targetItem);
              } catch (e) {
                // We can fail while creating the tooltip, since window 
              }
              // We're gonna paint it, so set it to false
              targetItem.showcaseCanvas.showcasePainted = false;
              if (paintThumbnailsAfterLoading) {
                // If it's a window's selected thumbnail, give it priority over the others
                if (!pendingThumbnails.contains(targetItem)) {
                  if (paintSelectedTabsFirst && (targetItem == currentSelectedTab)) {
                    pendingThumbnails.insertElementAt(targetItem, pendingSelectedThumbnailIndex);
                    pendingSelectedThumbnailIndex++;
                  } else {
                    pendingThumbnails.push(targetItem);
                  }
                }
              } else {
                setTimeout(paintThumbnail, 0, targetItem);
              }
            }
          }
        }
      }
    }
  }

  if (paintThumbnailsAfterLoading) {
    setTimeout(paintThumbnails, 0);
  }
}

function paintThumbnails() {
  if (pendingThumbnails.length > 0) {
    var targetItem = pendingThumbnails.shift();
    paintThumbnail(targetItem);
    setTimeout(paintThumbnails, 0);
  }
}

function paintThumbnail(targetThumbnail) {
  try {
  var showcaseCanvasBox = document.getBoxObjectFor(targetThumbnail.showcaseCanvas);
  var targetWidth = showcaseCanvasBox.width;
  var targetHeight = showcaseCanvasBox.height;
  
  targetThumbnail.showcaseCanvas.width = targetWidth;
  targetThumbnail.showcaseCanvas.height = targetHeight;

  if (targetThumbnail.showcaseIETab) {
    var ctx = targetThumbnail.showcaseCanvas.getContext("2d");
    ctx.clearRect(0, 0, targetWidth, targetHeight);
  } else {
    var ctx = targetThumbnail.showcaseCanvas.getContext("2d");
    ctx.clearRect(0, 0, targetWidth, targetHeight);
    ctx.save();
    ctx.scale(targetWidth/targetThumbnail.contentWindow.innerWidth, targetHeight/targetThumbnail.contentWindow.innerHeight);

    ctx.drawWindow(targetThumbnail.contentWindow, 
      targetThumbnail.contentWindow.scrollX, 
      targetThumbnail.contentWindow.scrollY, 
      targetThumbnail.contentWindow.innerWidth, 
      targetThumbnail.contentWindow.innerHeight, 
      "rgb(255,255,255)");
      ctx.restore();
  }
  } catch (e) {
    //Paint thumbnail failed, probably because the thumbnail disappeared. Fail silently.
  }

  // Mark it has been painted
  targetThumbnail.showcaseCanvas.showcasePainted = true;

  // Painting is a slow process, so we check if thumbnail has changed size while we were painting
  if ((targetWidth != showcaseCanvasBox.width) || (targetHeight != showcaseCanvasBox.height)) {
    // Hack to force resize
    setSize(targetThumbnail.showcaseCanvas, showcaseCanvasBox.width + 1, showcaseCanvasBox.height + 1);
    setTimeout(setSize, 0, targetThumbnail.showcaseCanvas, showcaseCanvasBox.width, showcaseCanvasBox.height);
  }
}

function updateTitle() {
  if (!showcaseOriginalTitle) {
    showcaseOriginalTitle = document.title;
  }

  var finalTitle = showcaseOriginalTitle + " [";

  if (globalMode) {
    if (showcaseWindows) {
      finalTitle += showcaseWindows.length;
      finalTitle += "/";
      var totalTabs = 0;
      for (var c=0; c<showcaseWindows.length; c++) {
        if (showcaseWindows[c].showcaseTabs) {
          totalTabs += showcaseWindows[c].showcaseTabs.length;
        }
      }
      finalTitle += totalTabs;
    } else {
      finalTitle += "?";
    }
  } else {
    if (showcaseWindows && (showcaseWindows.length > 0) && (showcaseWindows[0].showcaseTabs)) {
      finalTitle += showcaseWindows[0].showcaseTabs.length;
    } else {
      finalTitle += "?";
    }
  }

  finalTitle += "]";
  document.title = finalTitle;
}

function createTooltip(targetItem) {
  var tabTooltip = "";
  if (targetItem.showcaseProgressListener.progressListener.isLoading && !targetItem.showcaseCanvasBox.showcaseIsTitleLoaded) {
    if (!loadingTabTitle) {
      // Load locale string
      var tabbrowserBundle = document.getElementById("bundle_tabbrowser");
      loadingTabTitle= tabbrowserBundle.getString("tabs.loading");
    }
    tabTooltip = loadingTabTitle;
  } else if ((!targetItem.contentDocument) || (!targetItem.contentDocument.title) || (targetItem.contentDocument.title.length == 0)) {
    if (!untitledTabTitle) {
      // Load locale string
      var tabbrowserBundle = document.getElementById("bundle_tabbrowser");
      untitledTabTitle = tabbrowserBundle.getString("tabs.untitled");
    }
    tabTooltip = untitledTabTitle;
  } else {
    tabTooltip = targetItem.contentDocument.title;
  }

  var targetURL;
  if (targetItem.showcaseIETab) {
    try {
      targetURL = targetItem.contentDocument.URL.substring(IETAB_URL.length);
    } catch (stringException) {
      targetURL = targetItem.contentDocument.URL;
    }
  } else {
    targetURL = targetItem.contentDocument.URL;
  }

  if (thumbnailShowLabels) {
    if (targetItem.showcaseProgressListener.progressListener.isLoading) {
      targetItem.showcaseCanvasBox.showcaseLabel.setAttribute("value", tabTooltip);
      targetItem.showcaseCanvasBox.showcaseImageLabel.removeAttribute("src");
    } else {
      targetItem.showcaseCanvasBox.showcaseLabel.setAttribute("value", tabTooltip);
      if ((targetItem.mIconURL) && (targetItem.mIconURL != "")) {
        targetItem.showcaseCanvasBox.showcaseImageLabel.setAttribute("src", targetItem.mIconURL);
      } else {
        targetItem.showcaseCanvasBox.showcaseImageLabel.removeAttribute("src");
      }
    }
    targetItem.showcaseCanvas.setAttribute("tooltiptext", targetURL);
  } else {
    if (tabTooltip.length > 0) {
      tabTooltip += " - ";
    }
    tabTooltip += targetURL;
    targetItem.showcaseCanvas.setAttribute("tooltiptext", tabTooltip);
  }
}

function mosaicRefresh() {
  var thumbWidth = mosaicWidth / mosaicColumns;
  var thumbHeight = mosaicHeight / mosaicRows;
  var currentRow = 0;
  var currentColumn = 0;
  var c = 0;
  
  if (inZoomMode) {
    for (var i = 0; i < showcaseWindows.length; i++) {
      var targetWindow =  showcaseWindows[i];
      for (var j = 0; j<targetWindow.showcaseTabs.length; j++) {
        var targetTab = targetWindow.showcaseTabs[j];
        if (targetTab.showcaseCanvasBox) {
          // We just center them
          var yPosition = mosaicHeight / 2;
          var xPosition = mosaicWidth / 2;
          var topPosition = yPosition - targetTab.showcaseCanvasBox.height / 2;
          var leftPosition = xPosition - targetTab.showcaseCanvasBox.width / 2;
          targetTab.showcaseCanvasBox.setAttribute("top", topPosition);
          targetTab.showcaseCanvasBox.setAttribute("left", leftPosition);
          targetTab.showcaseCanvasBox.showcaseCenterX = xPosition;
          targetTab.showcaseCanvasBox.showcaseCenterY = yPosition;
        }
      }
    }
  } else {
    for (var i = 0; i < showcaseWindows.length; i++) {
      var targetWindow =  showcaseWindows[i];
      for (var j = 0; j<targetWindow.showcaseTabs.length; j++) {
        var targetTab = targetWindow.showcaseTabs[j];
        if (targetTab.contentWindow != window) {
          var yPosition = currentRow * thumbHeight + thumbHeight / 2;
          var xPosition = currentColumn * thumbWidth + thumbWidth / 2;
          var topPosition = yPosition - targetTab.showcaseCanvasBox.height / 2;
          var leftPosition = xPosition - targetTab.showcaseCanvasBox.width / 2;
          targetTab.showcaseCanvasBox.setAttribute("top", topPosition);
          targetTab.showcaseCanvasBox.setAttribute("left", leftPosition);
          targetTab.showcaseCanvasBox.showcaseCenterX = xPosition;
          targetTab.showcaseCanvasBox.showcaseCenterY = yPosition;
          targetTab.showcaseColumn = currentColumn;
          targetTab.showcaseRow = currentRow;

          currentColumn++;
          if (currentColumn >= mosaicColumns) {
            currentColumn = 0;
            currentRow++;
          }
        }
      }
    }
  }
}

function mosaicCalculateBestGrid(width, height) {
  var retValues = new Array(2);
  var numberRows, numberColumns;
  var currentArea, maxArea;
  var numTabs = 0;
  var maxTabWidth = -1;
  var maxTabHeight = -1;
  numberRows = 1;
  maxArea = -1;
  
  for (var i=0; i<showcaseWindows.length; i++) {
    var targetWindow = showcaseWindows[i];
    numTabs += targetWindow.showcaseTabs.length;
    if (targetWindow.showcaseLastGoodWidth > maxTabWidth) {
      maxTabWidth = targetWindow.showcaseLastGoodWidth;
    }
    if (targetWindow.showcaseLastGoodHeight > maxTabHeight) {
      maxTabHeight = targetWindow.showcaseLastGoodHeight;
    }
  }

  do {
    numberColumns = Math.floor(numTabs / numberRows);
    if ((numTabs % numberRows) > 0) {
      numberColumns++;
    }
    var thumbWidth = width / numberColumns;
    var thumbHeight = height / numberRows;

    if ((thumbWidth * maxTabHeight / maxTabWidth) > thumbHeight) {
      currentArea = thumbHeight * thumbHeight * maxTabWidth / maxTabHeight;
    } else {
      currentArea = thumbWidth * thumbWidth * maxTabHeight / maxTabWidth;
    }
    if (currentArea > maxArea) {
      retValues[0] = numberRows;
      retValues[1] = numberColumns;
      maxArea = currentArea;
    }
    numberRows++;
  } while (currentArea >= maxArea);

  return retValues;
}

// getTargetWindows
// Retrieves the target windows we want to "showcase"
function getTargetWindows() {
  var windowArray = new Array();
  if (globalMode) {
    // Target all windows
    var wm = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
    var windowIter = wm.getZOrderDOMWindowEnumerator('navigator:browser', true);
    var windowIndex = 0;
    
    while (windowIter.hasMoreElements()) {
      windowArray[windowIndex] = windowIter.getNext();
      windowIndex++;
    }
    
    if (windowIndex == 0) {
      // No windows?! Verify that with getEnumerator!
      var windowIter = wm.getEnumerator('navigator:browser');
    
      while (windowIter.hasMoreElements()) {
        windowArray[windowIndex] = windowIter.getNext();
        windowIndex++;
      }
    }
  } else {
    // Target "this" window (which is an argument)
    windowArray[0] = getParentWindow();
  }

  return windowArray;
}

function getParentWindow() {
  if (window.opener) {
    return window.opener;
  } else {
    return QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIWebNavigation).QueryInterface(Components.interfaces.nsIDocShellTreeItem).rootTreeItem.QueryInterface(Components.interfaces.nsIInterfaceRequestor).getInterface(Components.interfaces.nsIDOMWindow);
  }
}

function updateLoadingState(targetProgressTab) {
  targetProgressTab.showcaseCanvasBox.setAttribute("loading", targetProgressTab.showcaseProgressListener.progressListener.isLoading);
  targetProgressTab.showcaseCanvasBox.showcaseIsTitleLoaded = !targetProgressTab.showcaseProgressListener.progressListener.isLoading;
}

function updateIEState(targetProgressTab) {
  targetProgressTab.showcaseCanvasBox.setAttribute("ietab", targetProgressTab.showcaseIETab);
}

// Object TabProgressListener
function TabProgressListener(targetProgressTab ) {
  this.targetProgressTab = targetProgressTab;
  this.isLoading = false;
  this.progressListener = 
{
    QueryInterface: function(aIID)
    {
     if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
         aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
         aIID.equals(Components.interfaces.nsIXULBrowserWindow) ||
         aIID.equals(Components.interfaces.nsISupports))
       return this;
     throw Components.results.NS_NOINTERFACE;
    },

    onStateChange: function(aProgress, aRequest, aStateFlags, aStatus) {
     // Update thumbnail when something happens with the browser it represents
     if ((aStateFlags & STATE_START) && (aStateFlags & STATE_IS_WINDOW)) {
       this.isLoading = true;
       updateLoadingState(targetProgressTab);

       if (updateThumbnailWhenContentLoaded || updateThumbnailWhenContentChanges) {
         createTooltip(targetProgressTab);
       }
     } else if ((aStateFlags & STATE_STOP) && (aStateFlags & STATE_IS_WINDOW)) {
       if (targetProgressTab.contentDocument && targetProgressTab.contentDocument.URL) {
         var currentIETabState = (targetProgressTab.contentDocument.URL.substr(0, IETAB_URL.length) == IETAB_URL);
         if (currentIETabState != targetProgressTab.showcaseIETab) {
           targetProgressTab.showcaseIETab = currentIETabState;
           updateIEState(targetProgressTab);
           setTimeout(paintThumbnail, 0, targetProgressTab);
         }
         if (currentIETabState) {
           setTimeout(createTooltip, 0, targetProgressTab);
         }
       }

       this.isLoading = false;
       updateLoadingState(targetProgressTab);
       if (updateThumbnailWhenContentLoaded || updateThumbnailWhenContentChanges) {
         if (targetProgressTab.showcaseCanvas.showcasePainted) {
           setTimeout(createTooltip, 0, targetProgressTab);
           targetProgressTab.showcaseCanvas.showcasePainted = false;
           setTimeout(paintThumbnail, 0, targetProgressTab);
         }
       }
     }
     
     if (this.isLoading &&
              (!limitContentChangeUpdates || !targetProgressTab.showcaseTransferringLastUpdated || (targetProgressTab.showcaseTransferringLastUpdated < (new Date()).valueOf()))
       ) {
       if (targetProgressTab.showcaseCanvas.showcasePainted) {
         targetProgressTab.showcaseCanvas.showcasePainted = false;
         setTimeout(paintThumbnail, 0, targetProgressTab);
       }
       setTimeout(createTooltip, 0, targetProgressTab);
       targetProgressTab.showcaseTransferringLastUpdated = (new Date()).valueOf() + limitContentChangeUpdatesTime;
     }

     return 0;
    },

    onLocationChange: function(aProgress, aRequest, aURI)
    {
     // This fires when the location bar changes i.e load event is confirmed
     // or when the user switches tabs
     return 0;
    },

    // For definitions of the remaining functions see XulPlanet.com
    onProgressChange: function() {return 0;},
    onStatusChange: function() {return 0;},
    onSecurityChange: function() {return 0;},
    onLinkIconAvailable: function(icon) { alert("icon available. " + icon); return 0;}
  }
  
  this.targetProgressTab.addProgressListener(this.progressListener, Components.interfaces.nsIWebProgress.NOTIFY_ALL);
}

TabProgressListener.prototype.unloadListener = function(){
  try {
    this.targetProgressTab.removeProgressListener(this.progressListener);
  } catch (e) {}
}

// Object TabEventListener
function TabEventListener(targetTab, targetWindow) {
  this.listenTab = targetTab;
  this.tabwindow = targetWindow;
}

TabEventListener.prototype.handleEvent = function(evt){
  switch(evt.type){
    case 'resize': 
      // Update size
      try {
        this.tabwindow.showcaseLastGoodWidth = this.listenTab.contentWindow.innerWidth;
        this.tabwindow.showcaseLastGoodHeight = this.listenTab.contentWindow.innerHeight;
      } catch (e) {
        alert("tablisten1: " + e);
      }
      
      if (this.tabwindow.showcaseTabs) {
        for (var c=0; c<this.tabwindow.showcaseTabs.length; c++) {
          setTimeout(resizeTab, 0, this.tabwindow.showcaseTabs[c]);
        }
      } else {
        setTimeout(resizeTab, 0, this.listenTab);
      }
      break;
    case 'DOMTitleChanged':
       this.listenTab.showcaseCanvasBox.showcaseIsTitleLoaded = true;
       setTimeout(createTooltip, 0, this.listenTab);
       break;
//    case 'select':
//    case 'close':
//    case 'unload':
//      checkTabClosed(this.listenTab, this.tabbrowser);
//      setTimeout(checkTabClosed, 0, this.listenTab, this.tabwindow);
//      break;
  }
}

function resizeTab(targetTab) {
  updateThumbnail(targetTab, thumbWidth, thumbHeight, thumbnailIdleSizeFactor / 100, thumbnailActiveSizeFactor / 100);
  if (targetTab.showcaseCanvas.showcasePainted) {
    targetTab.showcaseCanvas.showcasePainted = false;
    paintThumbnail(targetTab);
  }

  mosaicRefresh();
}

// Object TabsChangeListener
function TabsChangeListener(targetWindow) {
  this.targetWindow = targetWindow;
}

TabsChangeListener.prototype.handleEvent = function(evt){
  switch(evt.type){
    case 'DOMNodeInserted':
      setTimeout(checkTabCreation, 0, this.targetWindow);
      break;
    case 'DOMNodeRemoved':
      setTimeout(checkTabDestruction, 0, this.targetWindow);
      break;
    case 'select':
      setTimeout(updateBorders, 0, this.targetWindow);
      break;
  }

//  checkTabCreation(this.targetWindow);
}

function checkTabCreation(targetWindow) {
  try {
  var found = false;
  var tabbrowser = targetWindow.document.getElementById("content");
  var showcaseBox = document.getElementById("showcase-box");

  for (var j = 0; j< tabbrowser.browsers.length; j++) {
    var targetTab = tabbrowser.browsers.item(j);
    if ((!targetTab.showcaseCanvasBox) && (targetTab.contentWindow != window)) {
      // link new tab
      if ((!targetTab.showcaseCanvasBox) && (targetTab.contentWindow != window)) {
        var newThumbnail = createThumbnail(targetTab, tabbrowser, targetWindow);
        showcaseBox.appendChild(newThumbnail);
      }
      // Set border, just in case this creation doesn't triggers a "select" event
      // (like "open link in tab"?)
      updateTabBorder(targetTab, targetWindow);
      targetWindow.showcaseTabs.push(targetTab);
      found = true;
    }
  }

  if (found) {
    updateTitle();
    refreshContent();
  }
  } catch (e) { alert("checkTabCreation - " + e); }
}

function checkTabDestruction(targetWindow) {
  try {
  var found = false;
  var tabbrowser = targetWindow.document.getElementById("content");
  var showcaseBox = document.getElementById("showcase-box");
  var tabArray = new Array();

  for (var j = 0; j< tabbrowser.browsers.length; j++) {
    tabArray[j] = tabbrowser.browsers.item(j);
  }
  var tabArrayArgument = new DualArgument(tabArray, targetWindow);

  if (targetWindow.showcaseTabs.some(singleTabDestructionCheck, tabArrayArgument)) {
    updateTitle();
    setTimeout(refreshContent, 0);
  }

  } catch (e) { alert("checkTabDestruction - " + e); }
}

// Helper function that will check if an element is inside another
function singleTabDestructionCheck(element, index, array) {
  if ((this.firstArgument.indexOf(element) == -1) && element.showcaseCanvasBox) {
    // Destruction found
    var showcaseBox = document.getElementById("showcase-box");
    showcaseBox.removeChild(element.showcaseCanvasBox);
    unlinkTab(element, this.secondArgument);
    return true;
  }

  return false;
}

// Helper object that simply holds an argument
function DualArgument(firstArgument, secondArgument) {
  this.firstArgument = firstArgument;
  this.secondArgument = secondArgument;
}

function showShowcaseOptions() {
  var features;
  try {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                          .getService(Components.interfaces.nsIPrefBranch);
    var instantApply = prefs.getBoolPref("browser.preferences.instantApply");
    features = "chrome,titlebar,toolbar,centerscreen," + (instantApply ? "dialog=no" : "modal");
  } catch (e) {
    features = "chrome,titlebar,toolbar,centerscreen,modal";
  }
  openDialog(OPTIONS_URL, "", features);
}